module: run-tests
synopsis: A demonstration of the test framework.
author: Matthias Hlzl
copyright: (c) 1999, Matthias Hlzl

define class <test-a> (<object>)
  slot slot-a :: <integer>,
    init-keyword: #"slot-a", init-value: 1;
end class <test-a>;

define class <test-b> (<object>)
  slot my-a :: <test-a>, required-init-keyword: #"my-a";
  slot slot-b, init-keyword: #"b";
end class <test-b>;

define method \= (lhs :: <test-b>, rhs :: <test-b>) => equal? :: <boolean>;
  lhs.my-a == lhs.my-a;
end method \=;

define method frob-a (a :: <test-a>) => frobbed-a :: <integer>;
  if (a.slot-a <= 0) a.slot-a := 1 end;
  a.slot-a;
end method frob-a;

define method frob-b (b :: <test-b>) => ();
  b.my-a.frob-a;
  b.slot-b := add(b.slot-b, b.my-a);
  b.my-a := make(<test-a>);
end method frob-b;

// Tests.
// =====

// We define three test suites, to show the framework.  In the first
// test-suite we put tests that should succeed, in the second one we
// put some tests that create failures as well.  The third test suite
// contains some tests with unexpected errors.

make(<test-suite>, id: #"empty test suite");
make(<test-suite>, id: #"empty tests");
make(<test-suite>, id: #"successful tests");
make(<test-suite>, id: #"one failure");
make(<test-suite>, id: #"failures");
make(<test-suite>, id: #"unexpected errors");
make(<test-suite>, id: #"nested tests");

for (i from 1 to 10)
  add-test(#"empty tests",
	   make(<empty-test-case>));
end for;

for (i from 1 to 10)
  add-test(#"empty tests",
	   make(<test-case>,
		description: "Empty test", test: method (test) end));
end for;

define class <a-test-case> (<test-case>)
  slot test-a1 :: <test-a>;
  slot test-a2 :: <test-a>;
  slot test-a3 :: <test-a>;
end class <a-test-case>;

define method set-up-test (a :: <a-test-case>) => ();
  a.test-a1 := make(<test-a>);
  a.test-a2 := make(<test-a>, slot-a: -1);
  a.test-a3 := make(<test-a>, slot-a: 0);
end method set-up-test;

add-test(#"successful tests",
	 make(<a-test-case>,
	      description: "Test instance creation",
	      test: method (test-case)
		      expect-instance(test-case.test-a1, <test-a>);
		      expect-instance(test-case.test-a2, <test-a>);
		      expect-instance(test-case.test-a3, <test-a>);
		    end method));

// We don't actually use the runner, it is just passed in to show that
// it can be done.
//
define function test-frob-a
    (test-case :: <a-test-case>, runner :: <test-runner>) => ();
  expect(frob-a(test-case.test-a1), 0, predicate: \>);
  expect(frob-a(test-case.test-a2), 0, predicate: \>);
  expect(frob-a(test-case.test-a3), 0, predicate: \>); 
end function test-frob-a;

add-test(#"successful tests",
	 make(<a-test-case>,
	      description: "Test frob-a",
	      test: test-frob-a));

define method test-frob-a-with-failure
    (test-case :: <a-test-case>) => ();
  expect(test-case.test-a1.frob-a, 0, predicate: \<);
end method test-frob-a-with-failure;

add-test(#"one failure",
	 make(<a-test-case>,
	      description: "Test instance creation and fail",
	      test: method (test-case, runner)
		      expect-instance(test-case.test-a1, <test-b>);
		    end method));

add-test(#"failures",
	 make(<a-test-case>,
	      description: "Test instance creation",
	      test: method (test-case, runner)
		      expect-instance(test-case.test-a1, <test-a>);
		      expect-instance(test-case.test-a2, <test-a>);
		      expect-instance(test-case.test-a3, <test-a>);
		    end method));
add-test(#"failures",
	 make(<a-test-case>,
	      description: "Test instance creation and fail",
	      test: method (test-case, runner)
		      expect-instance(test-case.test-a1, <test-b>);
		    end method));
add-test(#"failures",
	 make(<a-test-case>,
	      description: "Test frob-a",
	      test: test-frob-a));
add-test(#"failures",
	 make(<a-test-case>,
	      description: "Test frob a and fail",
	      test: test-frob-a-with-failure));

define method test-frob-a-with-unexpected-error
    (test-case :: <a-test-case>) => ();
  expect(test-case.test-a1.frob-a, 0, predicate: \>);
  error("Failing unexpectedly.");
end method test-frob-a-with-unexpected-error;

add-test(#"unexpected errors",
	 make(<a-test-case>,
	      description: "Test instance creation and fail",
	      test: method (test-case)
		      error("Failing unexpectedly.");
		    end method));
add-test(#"unexpected errors",
	 make(<a-test-case>,
	      description: "Test frob-a",
	      test: test-frob-a));
add-test(#"unexpected errors",
	 make(<a-test-case>,
	      description: "Test frob-a and fail",
	      test: test-frob-a-with-failure));
add-test(#"unexpected errors",
	 make(<a-test-case>,
	      description: "Test frob-a and fail unexpectedly",
	      test: test-frob-a-with-unexpected-error));

add-test(#"nested tests",
	 make(<test-case>,
	      description: "First empty test", test: method (t, r) end));
add-test(#"nested tests", test-suite(#"empty tests"));
add-test(#"nested tests",
	 make(<test-case>,
	      description: "Second empty test", test: method (t, r) end));
add-test(#"nested tests", test-suite(#"empty test suite"));
add-test(#"nested tests", test-suite(#"successful tests"));

define class <b-test-case> (<test-case>)
  inherited slot test-case-description = "b-test-case";
  slot test-a :: <test-a> = make(<test-a>);
  slot test-b :: <test-b> = make(<test-b>, my-a: make(<test-a>));
end class <b-test-case>;

add-test(#"unexpected errors",
	 make(<b-test-case>,
	      description: "Test uninitialized slot access",
	      test: method (test) expect(test.test-b.slot-b, 0) end));

define method main(appname, #rest arguments)
  format-out("\nTesting the test-framework.\n");
  format-out("==========================\n\n");
  
  run-all-tests(#"empty test suite");
  run-all-tests(#"empty tests");
  run-all-tests(#"successful tests");
  run-all-tests(#"one failure");
  run-all-tests(#"failures");
  run-all-tests(#"this test suite does not exist");
  run-all-tests(#"unexpected errors");
  run-all-tests(#"nested tests");

  format-out("\n**********************************************\n");
  
  run-all-known-tests();
  
  exit(exit-code: 0);
end method main;
