module: test-framework-internals
synopsis: "A simple test framework, similar to Kent Beck's Smalltalk"
          "test framework or JUnit."
author: Matthias Hlzl
copyright: (c) 1999, Matthias Hlzl

// Reflection support.

// These classes are copied from the reflection module of the
// utilities library.  This is done to avoid circular dependencies
// between the utilities and test-framework libraries (after all we
// want to write a test suite for the utilities).

// all-subclasses
//
// Returns all subclasses of class that we can know about.  If class
// or one of its subclasses is sealed this may not be all subclasses
// of class.
//
define method all-subclasses (class :: <class>)
 => (subclasses :: <sequence>, unknown-subclasses :: <boolean>);
  // format-out("all-subclasses{<class>}: %=", class);
  *standard-output*.force-output;
  block ()
    all-subclasses(direct-subclasses(class));
  exception (<sealed-object-error>)
    values(#(), #t);
  end block;
end method all-subclasses;

define method all-subclasses (classes :: <sequence>)
  => (subclasses :: <sequence>, unknown-subclasses :: <boolean>);
  // format-out("all-subclasses{<sequence>}: %=\n", classes);
  // *standard-output*.force-output;
  let unknown-subclasses :: <boolean> = #f;
  let subclasses :: <sequence> = #[];
  for (c in classes)
    let new-subclasses = calculate-new-subclasses(c, classes);
    if (new-subclasses)
      subclasses := union(subclasses, new-subclasses);
    else
      unknown-subclasses := #t;
    end if;
  end for;
  if (subclasses.empty?)
    values(classes, unknown-subclasses);
  else
    all-subclasses(union(classes, subclasses));
  end if;
end method all-subclasses;

define function calculate-new-subclasses
    (class :: <class>, known-subclasses :: <sequence>)
 => new-subclasses :: false-or(<sequence>);
  block ()
    let subclasses = direct-subclasses(class);
    for (duplicate in known-subclasses)
      // format-out("  removing %=\n", duplicate);
      subclasses := remove(subclasses, duplicate);
    end for;
    // format-out("  adding %=\n", subclasses);
    subclasses;
  exception (<sealed-object-error>)
    #f;
  end block;
end function calculate-new-subclasses;

// The class <test> is the root of a composite pattern.
//
define abstract class <test> (<object>)
end class <test>;

// A raw test case.  Subclasses of <test-case> contain the fixtures
// and code to run a single test.

define abstract open class <general-test-case> (<test>)
end class <general-test-case>;

define class <empty-test-case> (<general-test-case>)
end class <empty-test-case>;

define open class <test-case> (<general-test-case>)
  constant slot test-case-description :: <string>,
    required-init-keyword: #"description";
  constant slot test-case-test :: <function>,
    required-init-keyword: #"test";
end class <test-case>;

define variable *current-test-case* :: <general-test-case>
  = make(<empty-test-case>);

// Test suites are collections of <test>s.
//

define open class <test-suite> (<test>)
  slot test-suite-id :: <symbol>,
    required-init-keyword: #"id";
  slot tests :: <deque>,
    init-keyword: #"tests", init-function: curry(make, <deque>);
end class <test-suite>;

define variable *test-suites* :: <table> = make(<table>);

define function test-suites() => all-test-suites :: <table>;
  *test-suites*;
end function test-suites;

define function test-suite
    (id :: <symbol>, #key default :: false-or(<test-suite>) = #f)
 => test-suite :: false-or(<test-suite>);
    element(*test-suites*, id, default: #f);
end function test-suite;

define function test-suite-setter
    (new-test-suite :: <test-suite>, id :: <symbol>)
 => old-test-suite-with-same-key :: false-or(<test-suite>);
  let old-test-suite = test-suite(id);
  *test-suites*[id] := new-test-suite;
  old-test-suite;
end function test-suite-setter;

define method initialize(self :: <test-suite>, #key)
  next-method();
  *test-suites*[self.test-suite-id] := self;
end method initialize;

// Initialization of test suites.

// Initialization for the test.  Maybe this should be done in an
// initialize method?
//
define open generic set-up-test (test :: <test>) => ();

define method set-up-test (test :: <test>) => ();
  // Yup, do nothing.
end method set-up-test;

// Finalization for the test.
//
define open generic tear-down-test (test :: <test>) => ();

define method tear-down-test (test :: <test>) => ();
  // How interesting...
end method tear-down-test;

// Adding a new test.
//
define open generic add-test
    (test-suite-designator :: type-union(<test-suite>, <symbol>),
     test-case :: <test>) => ();

define method add-test
    (test-suite :: <test-suite>, test :: <test>) => ();
  push-last(test-suite.tests, test);
end method add-test;

define method add-test
    (name :: <symbol>, test-case :: <test>) => ();
  let suite = test-suite(name);
  if (suite)
    add-test(suite, test-case);
  else
    format-out("add-test: No test case named %=.\n", name);
    *standard-output*.force-output;
  end if;
end method add-test;

// Handling errors in tests.
//
define open class <test-failure> (<serious-condition>)
  constant slot runner :: <test-runner>,
    required-init-keyword: #"runner";
  constant slot test-case :: <general-test-case>,
    required-init-keyword: #"test-case";
end class <test-failure>;

define open generic fail-test (test-case :: <general-test-case>) => ();

define method fail-test (test-case :: <test-case>) => ();
  let failure = make(<test-failure>, runner: *current-runner*,
		     test-case: test-case);
  signal(failure);
end method fail-test;

define function expect
    (test-result, expected-value, #key predicate :: <function> = \==)
 => ();
  expect-aux(test-result, expected-value, predicate, *current-test-case*);
end function expect;

define method expect-aux
    (test-result, expected-value, predicate :: <function>,
     test-case :: <test-case>) => ();
  unless (predicate(test-result, expected-value))
    fail-test(*current-test-case*);
  end unless;
end method expect-aux;

define method expect-aux
    (test-result, expected-value, predicate :: <function>,
     test-case :: <empty-test-case>) => ();
  format-out("expect:empty-test-case?");
  *standard-output*.force-output;
end method expect-aux;

// Expect-value is like `expect' but it should create a more
// specialized condition in case of failure.
//
define function expect-value (test-result, expected-value) => ();
  expect(test-result, expected-value);
end function expect-value;

define function expect-instance (test-result, expected-type) => ();
  expect(test-result, expected-type, predicate: instance?);
end function expect-instance;

define function note-failure
    (failure :: type-union(<test-failure>, <unexpected-failure>),
     runner :: <test-runner>)
 => ();
  runner.failures := runner.failures + 1;
  push-last(runner.failed-tests, failure);
end function note-failure;

define open generic register-failed-test (failure :: <test-failure>) => ();

define method register-failed-test (failure :: <test-failure>) => ();
  note-failure(failure, failure.runner)
end method register-failed-test;

// Handling unexpected errors during tests.

// Instances of <unexpected-failure> are used to signal errors that we
// didn't account for in the test-suite.
//
define open class <unexpected-failure> (<error>)
  slot original-exception :: <condition>,
    required-init-keyword: #"original-exception";
end class <unexpected-failure>;

define open generic register-unexpected-failure
    (failure :: <unexpected-failure>) => ();

define method register-unexpected-failure
    (failure :: <unexpected-failure>) => ();
  note-failure(failure, *current-runner*);
end method register-unexpected-failure;

// A <test-runner> is used to run a test-suite.
//
define open class <general-test-runner> (<object>)
end class <general-test-runner>;

define class <empty-test-runner> (<general-test-runner>)
end class <empty-test-runner>;

define open class <test-runner> (<general-test-runner>)
  constant slot runner-test-suite :: <test-suite>,
    required-init-keyword: #"test-suite";
  slot current-test :: limited(<integer>, min: 0),
    init-value: 0;
  slot failures :: limited(<integer>, min: 0),
    init-value: 0;
  slot failed-tests :: <deque> = make(<deque>);
end class <test-runner>;

define variable *current-runner* :: <general-test-runner>
  = make(<empty-test-runner>);
  
define open generic run-all-tests (test-or-runner :: <object>) => ();

define method run-all-tests (runner :: <empty-test-runner>) => ();
  // What do you think an empty runner is supposed to do?
  // Right.
end method run-all-tests;

define method run-all-tests (runner :: <test-runner>) => ();
  *current-runner* := runner;
  run-tests(runner.runner-test-suite, runner);
  runner.print-test-summary;
end method run-all-tests;

define method run-all-tests (suite :: <test-suite>) => ();
  let runner = make(<test-runner>, test-suite: suite);
  format-out("*** Running test-suite %= ***\n", suite.test-suite-id);
  *standard-output*.force-output;
  run-all-tests(runner);
  format-out("--- Done with test-suite %= ---\n", suite.test-suite-id);
  *standard-output*.force-output;
end method run-all-tests;

define method run-all-tests (suite-id :: <symbol>) => ();
  let suite = test-suite(suite-id);
  if (suite)
    run-all-tests(suite);
  else
    format-out("run-all-tests: No test suite named %=.\n", suite-id);
    *standard-output*.force-output;
  end if;
end method run-all-tests;

define function run-all-known-tests () => ();
  // format-out("run-all-known-tests\n");
  for (test in test-suites())
    run-all-tests(test);
  end for;
end function run-all-known-tests;

// Running a test case.
//
define open generic run-test-case
    (test-case :: <general-test-case>, runner :: <test-runner>) => ();

define method run-test-case
    (test-case :: <empty-test-case>, runner :: <test-runner>) => ();
  format-out(".");
  *standard-output*.force-output;
end method run-test-case;

define method run-test-case
    (test-case :: <test-case>, runner :: <test-runner>) => ();
  *current-test-case* := test-case;
  block ()
    test-case.set-up-test;
    let test-fun = test-case.test-case-test;
    if (test-fun.function-specializers.size = 2)
      test-fun(test-case, runner);
    else
      test-fun(test-case);
    end if;
    format-out(".");
  cleanup
    test-case.tear-down-test;
    force-output(*standard-output*);
  exception (failure :: <test-failure>)
    register-failed-test(failure);
    format-out("F");
  exception (failure :: <condition>)
    let new-failure = make(<unexpected-failure>, original-exception: failure);
    register-unexpected-failure(new-failure);
    format-out("E");
  end block;
  *standard-output*.force-output;
end method run-test-case;

// Running tests.
//
define open generic run-tests
    (test :: <test>, runner :: <test-runner>) => ();

define method run-tests
    (test :: <general-test-case>, runner :: <test-runner>) => ();
  run-test-case(test, runner);
end method run-tests;

define method run-tests
    (test :: <test-suite>, runner :: <test-runner>) => ();
  for (t in test.tests)
    run-tests(t, runner);
  end for;
end method run-tests;

// Printing.
//

define open generic print-test-summary (runner :: <test-runner>) => ();

define method print-test-summary (runner :: <test-runner>) => ();
  format-out("\n");
  if (runner.failures = 0)
    format-out(">>> All tests completed without errors. <<<\n");
  elseif (runner.failures = 1)
    format-out("!!! There was one failure !!!\n");
    runner.print-runner-failures;
  else
    format-out("!!! There were %D failures !!!\n", runner.failures);
    runner.print-runner-failures;
  end if;
  *standard-output*.force-output;
end method print-test-summary;

define open generic print-runner-failures (runner :: <test-runner>) => ();

define method print-runner-failures (runner :: <test-runner>) => ();
  for (failure in runner.failed-tests)
    failure.print-failure;
  end for;
end method print-runner-failures;

define open generic print-failure (failure :: <condition>) => ();

define method print-failure (failure :: <test-failure>) => ();
  format-out("  %S failed.\n", failure.test-case.test-case-description);
end method print-failure;

define method print-failure (failure :: <unexpected-failure>) => ();
  let original = failure.original-exception;
  if (instance?(original, <uninitialized-slot-error>))
    format-out("  Unexpected failure:\n");
    format-out("    uninitialized slot %=\n",
	       original.uninitialized-slot-error-slot.slot-name);
    format-out("    in instance %=\n",
	       original.uninitialized-slot-error-instance);
  else
    format-out("  Unexpected failure: %=!\n", original);
  end if;
  *standard-output*.force-output;
end method print-failure;

/*
define method print-failure (failure :: <unexpected-failure>) => ();
  let original = failure.original-exception;
  format-out("  Unexpected failure: %=!\n", original);
  *standard-output*.force-output;
end method print-failure;
*/